=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2007-02-26 21:17:03 +0000
+++ bzrlib/dirstate.py	2007-02-26 22:08:14 +0000
@@ -320,6 +320,13 @@
         dirname, basename = osutils.split(utf8path)
         assert file_id.__class__ == str, \
             "must be a utf8 file_id not %s" % (type(file_id))
+        # Make sure the file_id does not exist in this tree
+        file_id_entry = self._get_entry(0, fileid_utf8=file_id)
+        if file_id_entry != (None, None):
+            path = osutils.pathjoin(file_id_entry[0][0], file_id_entry[0][1])
+            kind = DirState._minikind_to_kind[file_id_entry[1][0][0]]
+            info = '%s:%s' % (kind, path)
+            raise errors.DuplicateFileId(file_id, info)
         entry_key = (dirname, basename, file_id)
         self._read_dirblocks_if_needed()
         block_index, present = self._find_block_index_from_key(entry_key)

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2007-02-25 22:50:34 +0000
+++ bzrlib/errors.py	2007-02-26 22:08:14 +0000
@@ -1377,6 +1377,16 @@
         self.file_id = file_id
 
 
+class DuplicateFileId(BzrError):
+
+    _fmt = "File id {%(file_id)s} already exists in inventory as %(entry)s"
+
+    def __init__(self, file_id, entry):
+        BzrError.__init__(self)
+        self.file_id = file_id
+        self.entry = entry
+
+
 class DuplicateKey(BzrError):
 
     _fmt = "Key %(key)s is already present in map"

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2007-02-22 15:04:35 +0000
+++ bzrlib/inventory.py	2007-02-26 22:08:14 +0000
@@ -1095,7 +1095,8 @@
         Returns the new entry object.
         """
         if entry.file_id in self._byid:
-            raise BzrError("inventory already contains entry with id {%s}" % entry.file_id)
+            raise errors.DuplicateFileId(entry.file_id,
+                                         self._byid[entry.file_id])
 
         if entry.parent_id is None:
             assert self.root is None and len(self._byid) == 0

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2007-02-25 22:50:34 +0000
+++ bzrlib/tests/test_errors.py	2007-02-26 22:08:14 +0000
@@ -30,6 +30,11 @@
 
 class TestErrors(TestCaseWithTransport):
 
+    def test_duplicate_file_id(self):
+        error = errors.DuplicateFileId('a_file_id', 'foo')
+        self.assertEqualDiff('File id {a_file_id} already exists in inventory'
+                             ' as foo', str(error))
+
     def test_inventory_modified(self):
         error = errors.InventoryModified("a tree to be repred")
         self.assertEqualDiff("The current inventory for the tree 'a tree to "

=== modified file 'bzrlib/tests/workingtree_implementations/test_add.py'
--- bzrlib/tests/workingtree_implementations/test_add.py	2007-02-26 16:53:16 +0000
+++ bzrlib/tests/workingtree_implementations/test_add.py	2007-02-26 22:08:14 +0000
@@ -51,6 +51,17 @@
 
         self.assertTreeLayout([('', root_id), ('one', 'one-id')], tree)
 
+    def test_add_existing_id(self):
+        """Adding an entry with a pre-existing id raises DuplicateFileId"""
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['a', 'b'])
+        tree.add(['a'], ['an-id'])
+        self.assertRaises(errors.DuplicateFileId,
+                          tree.add, ['b'], ['an-id'])
+        root_id = tree.get_root_id()
+        # And the entry should not have been added.
+        self.assertTreeLayout([('', root_id), ('a', 'an-id')], tree)
+
     def test_add_one_list(self):
         tree = self.make_branch_and_tree('.')
         self.build_tree(['one'])

